#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptcat.h"

#include <cptutils/cpt-write.h>
#include <cptutils/cpt-read.h>
#include <cptutils/btrace.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

typedef struct
{
  cpt_t *cpt;
  const char *file;
  double range[2];
} cptdat_t;

static int datcmp(const void *va, const void *vb)
{
  const cptdat_t *a = va, *b = vb;
  if (a->range[0] < b->range[0])
    return -1;
  if (a->range[0] > b->range[0])
    return 1;
  return 0;
}

static int cptcat2(cptcat_opt_t *opt, cpt_t *cpt, cptdat_t *dat)
{
  size_t n = opt->input.n;

  /* check all have the same model */

  model_t model = dat[0].cpt->model;

  for (size_t i = 1 ; i < n ; i++)
    {
      if (dat[0].cpt->model != model)
	{
	  btrace("incompatible colour models in input");
	  return 1;
	}
    }

  cpt->model = model;

  cpt->nan = dat[0].cpt->nan;
  cpt->bg = dat[0].cpt->bg;
  cpt->fg = dat[n-1].cpt->fg;

  for (size_t i = 0 ; i < n ; i++)
    {
      cpt_seg_t *seg;

      while ((seg = cpt_pop(dat[i].cpt)) != NULL)
	{
	  if (cpt_append(seg, cpt) != 0)
	    {
	      btrace("failed prepend");
	      return 1;
	    }
	}
    }

  /*
    create hyphenated name for new file, this has length which is
    the sum of lengths of the substrings, plus n - 1  hyphens,
    plus a termating null.  If n were to be zero then the malloc
    would be zero and the first strcpy would be out-of-bounds.
    In fact, n cannot be zero, that is ensured in main.c, but gcc-11
    warns of out-of-bounds, so we assert to suppress that warning.
  */

  assert(n > 0);

  size_t nctot = 0;

  for (size_t i = 0 ; i < n ; i++)
    nctot += (dat[i].cpt->name ? strlen(dat[i].cpt->name) : 1);

  cpt->name = malloc(nctot + n);

  strcpy(cpt->name, (dat[0].cpt->name ? dat[0].cpt->name : "x"));

  for (size_t i = 1 ; i < n ; i++)
    {
      strcat(cpt->name, "-");
      strcat(cpt->name, (dat[i].cpt->name ? dat[i].cpt->name : "x"));
    }

  /* destroy components */

  for (size_t i = 0 ; i < n ; i++)
    cpt_destroy(dat[i].cpt);

  /* write to file*/

  if (cpt_write(cpt, opt->output) != 0)
    {
      btrace("failed write to %s", opt->output);
      return 1;
    }

  return 0;
}

int cptcat(cptcat_opt_t opt)
{
  size_t n = opt.input.n;
  cptdat_t dat[n];

  /* load cpts */

  for (size_t i = 0 ; i < n ; i++)
    {
      dat[i].file = opt.input.file[i];

      if ((dat[i].cpt = cpt_read(dat[i].file)) == NULL)
	{
	  btrace("failed to read %s", dat[i].file);
	  return 1;
	}

      if (cpt_zrange(dat[i].cpt, dat[i].range) != 0)
	{
	  btrace("failed to get zrange");
	  return 1;
	}

      if (opt.verbose)
	printf("  %s\n", dat[i].file);
    }

  /*
    check increasing -- we could coerce, but that can
    can be done by applying makecpt to the offending
    component file.
  */

  for (size_t i = 0 ; i < n ; i++)
    {
      double *range = dat[i].range;

      if (range[0] >= range[1])
	{
	  btrace("%s is decreasing", dat[i].file);
	  return 1;
	}
    }

  /* sort */

  qsort(dat, n, sizeof(cptdat_t), datcmp);

  /* check contiguous */

  for (size_t i = 0 ; i + 1 < n ; i++)
    {
      if (dat[i].range[1] != dat[i+1].range[0])
	{
	  btrace("non-contiguous input:");
	  btrace("  %s : %g < z < %g",
		     dat[i].file, dat[i].range[0], dat[i].range[1]);
	  btrace("  %s : %g < z < %g",
		     dat[i+1].file, dat[i+1].range[0], dat[i+1].range[1]);
	  return 1;
	}
    }

  /* create the new cpt */

  cpt_t *cpt;

  if ((cpt = cpt_new()) == NULL)
    return 1;

  int err = cptcat2(&opt, cpt, dat);

  cpt_destroy(cpt);

  return err;
}
