#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptclip.h"

#include <cptutils/cpt-read.h>
#include <cptutils/cpt-write.h>
#include <cptutils/btrace.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

static int cptclip_convert(cpt_t*, cptclip_opt_t);

int cptclip(char *infile, char *outfile, cptclip_opt_t opt)
{
  cpt_t *cpt;
  int err = 1;

  if ((cpt = cpt_read(infile)) != NULL)
    {
      if (cptclip_convert(cpt, opt) == 0)
        {
          if (cpt_nseg(cpt) > 0)
            {
              if (cpt_write(cpt, outfile) == 0)
                err = 0;
              else
                btrace("error writing cpt struct");
            }
          else
            btrace("clipped result has no segments");

        }
      else
        btrace("failed to convert");

      cpt_destroy(cpt);
    }
  else
    btrace("failed to load cpt from %s", (infile ? infile : "<stdin>"));

  if (err)
    btrace("failed to write cpt to %s",
	   (outfile ? outfile : "<stdout>"));

  return err;
}

static int cptclip_z_inc(cpt_t*, cptclip_opt_t);
static int cptclip_z_dec(cpt_t*, cptclip_opt_t);

static int cptclip_convert(cpt_t *cpt, cptclip_opt_t opt)
{
  if (cpt->segment == NULL)
    {
      btrace("cpt has no segments");
      return 1;
    }

  if (opt.segments)
    {
      int nseg = cpt_nseg(cpt);

      for (int i = 1 ; i < opt.segs.min ; i++)
	{
	  cpt_seg_t *s = cpt_pop(cpt);
	  cpt_seg_destroy(s);
	}

      for (int i = nseg ; i > opt.segs.max ; i--)
	{
	  cpt_seg_t *s = cpt_shift(cpt);
	  cpt_seg_destroy(s);
	}

      return 0;
    }
  else
    {
      int err;

      if (cpt_increasing(cpt))
        err = cptclip_z_inc(cpt, opt);
      else
        err = cptclip_z_dec(cpt, opt);

      return err;
    }
}

/*
  handles cptclip_z_inc and _dec by considering the target
  interval in the gradient's left-right frame of reference.
*/

static int cptclip_z_id(cpt_t *cpt,
			double zleft,
			double zright,
			bool (*left_of)(double, double),
			bool (*right_of)(double, double))
{
  cpt_seg_t *seg;

  /*
    pop segments from the left until one of them
    intersects the requested interval
  */

  while ((seg = cpt_pop(cpt)) != NULL)
    {
      if (right_of(seg->rsmp.val, zleft))
	break;

      cpt_seg_destroy(seg);
    }

  if (seg)
    {
      /*
	if this segment is not completely inside the
	requested interval then clip it to the interval
      */

      if (left_of(seg->lsmp.val, zleft))
	{
	  cpt_fill_t fill;
	  double z =
	    (zleft - seg->lsmp.val) /
	    (seg->rsmp.val - seg->lsmp.val);

	  if  (cpt_fill_interpolate(z,
                                    seg->lsmp.fill,
                                    seg->rsmp.fill,
                                    cpt->model,
                                    &fill) != 0)
            {
              btrace("fill interpolation failed");
              return 1;
            }

	  seg->lsmp.fill = fill;
	  seg->lsmp.val = zleft;
	}

      /*
	push the (possibly modified) segment back to
	where it came
      */

      cpt_prepend(seg, cpt);
    }

  /* likewise on the right */

  while ((seg = cpt_shift(cpt)) != NULL)
    {
      if (left_of(seg->lsmp.val, zright))
	break;

      cpt_seg_destroy(seg);
    }

  if (seg)
    {
      if (right_of(seg->rsmp.val, zright))
	{
	  cpt_fill_t fill;
	  double z =
	    (zright - seg->lsmp.val) /
	    (seg->rsmp.val - seg->lsmp.val);

	  if  (cpt_fill_interpolate(z,
                                    seg->lsmp.fill,
                                    seg->rsmp.fill,
                                    cpt->model,
                                    &fill) != 0)
            {
              btrace("fill interpolation failed");
              return 1;
            }

	  seg->rsmp.fill = fill;
	  seg->rsmp.val = zright;
	}

      cpt_append(seg, cpt);
    }

  return 0;
}

static bool lt(double a, double b){ return a < b; }
static bool gt(double a, double b){ return a > b; }

static int cptclip_z_inc(cpt_t *cpt, cptclip_opt_t opt)
{
  return cptclip_z_id(cpt, opt.z.min, opt.z.max, lt, gt);
}

static int cptclip_z_dec(cpt_t *cpt, cptclip_opt_t opt)
{
  return cptclip_z_id(cpt, opt.z.max, opt.z.min, gt, lt);
}
