#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "options.h"
#include "cptclip.h"

#include <cptutils/btrace.h>

#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

int main(int argc, char **argv)
{
  struct gengetopt_args_info info;
  char *infile = NULL, *outfile = NULL;
  cptclip_opt_t opt = {0};

  /* use gengetopt */

  if (options(argc, argv, &info) != 0)
    {
      fprintf(stderr,"failed to parse command line\n");
      return EXIT_FAILURE;
    }

  /* check arguments & transfer to opt structure */

  opt.verbose = info.verbose_given;

  /* null outfile for stdout */

  if (info.output_given)
    outfile = info.output_arg;
  else
    {
      if (opt.verbose)
        {
          fprintf(stderr, "no --verbose without --output\n");
          return EXIT_FAILURE;
        }
      outfile = NULL;
    }

  /* say hello */

  if (opt.verbose)
    printf("This is cptclip (version %s)\n",VERSION);

  /* null infile for stdin */

  switch (info.inputs_num)
    {
    case 0:
      infile = NULL;
      break;

    case 1:
      infile = info.inputs[0];
      break;

    default:
      fprintf(stderr,"expected 0 or 1 arguments, got %i\n",
	      info.inputs_num);
      return EXIT_FAILURE;
    }

  /* segments option */

  opt.segments = info.segments_given;

  if (opt.segments)
    {
      if (sscanf(info.range_arg,"%zu/%zu",
		 &opt.segs.min,
		 &opt.segs.max) != 2)
	{
	  fprintf(stderr,"bad range argument %s\n",info.range_arg);
	  return EXIT_FAILURE;
	}

      if ((opt.segs.min > opt.segs.max) ||
	  ( opt.segs.min < 1 ) ||
	  ( opt.segs.max < 1 ))
	{
	  fprintf(stderr,"bad segment selection %zi - %zi\n",
		  opt.segs.min,opt.segs.max);
	  return EXIT_FAILURE;
	}

      if (opt.verbose)
	{
	  printf("extracting segments %zi to %zi\n",
		 opt.segs.min,opt.segs.max);
	}
    }
  else
    {
      if (sscanf(info.range_arg,"%lf/%lf",
		 &opt.z.min,
		 &opt.z.max) != 2)
	{
	  fprintf(stderr,"bad range argument %s\n",info.range_arg);
	  return EXIT_FAILURE;
	}

      if (opt.z.min >= opt.z.max)
	{
	  fprintf(stderr,"bad range %g - %g\n",
		  opt.z.min, opt.z.max);
	  return EXIT_FAILURE;
	}

      if (opt.verbose)
	{
	  printf("extracting z-range %g to %g\n",
		 opt.z.min, opt.z.max);
	}

    }

  btrace_enable("cptclip");

  int err = cptclip(infile,outfile,opt);

  if (err)
    {
      btrace("failed (error %i)", err);
      btrace_print_stream(stderr, BTRACE_PLAIN);

      if (info.backtrace_file_given)
        {
          int format = BTRACE_XML;
          if (info.backtrace_format_given)
            {
              format = btrace_format(info.backtrace_format_arg);
              if (format == BTRACE_ERROR)
                {
                  fprintf(stderr, "no such backtrace format %s\n",
                          info.backtrace_format_arg);
                  return EXIT_FAILURE;
                }
            }
          btrace_print(info.backtrace_file_arg, format);
        }
    }
  else
    {
      if (opt.verbose)
        printf("gradient written to %s\n",(outfile ? outfile : "<stdin>"));
    }

  btrace_reset();
  btrace_disable();

  if (opt.verbose)
    printf("done.\n");

  options_free(&info);

  return (err ? EXIT_FAILURE : EXIT_SUCCESS);
}
