#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptsvg.h"

#include <cptutils/btrace.h>
#include <cptutils/cpt-read.h>
#include <cptutils/svg-write.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

static int cptsvg_convert(cpt_t*, svg_t*, cptsvg_opt_t);

int cptsvg(char *infile, char *outfile, cptsvg_opt_t opt)
{
  cpt_t *cpt;
  int err = 1;

  if ((cpt = cpt_read(infile)) != NULL)
    {
      svg_t *svg;

      if ((svg = svg_new()) != NULL)
        {
          if (cptsvg_convert(cpt, svg, opt) == 0)
            {
              if (svg_write((const svg_t**)(&svg),
                            1,
                            outfile,
                            &(opt.preview)) == 0)
                err = 0;
              else
                btrace("error writing sgv struct");
            }
          else
            btrace("failed to convert cpt to svg");

          svg_destroy(svg);
        }
      else
        btrace("failed to allocate svg");

      cpt_destroy(cpt);
    }
  else
    btrace("failed to load cpt from %s", (infile ? infile : "<stdin>"));

  if (err)
    btrace("failed to write svg to %s", (outfile ? outfile : "<stdout>"));

  return err;
}

static int cptsvg_convert(cpt_t *cpt, svg_t *svg, cptsvg_opt_t opt)
{
  cpt_seg_t *seg;
  svg_stop_t lstop, rstop;
  double min, max;
  int n, m = 0;

  /* check we have at least one cpt segment */

  if (cpt->segment == NULL)
    {
      btrace("cpt has no segments");
      return 1;
    }

  /* copy the name */

  if (cpt->name)
    {
      if (snprintf((char*)svg->name,
		   SVG_NAME_LEN,
		   "%s", cpt->name) >= SVG_NAME_LEN)
	btrace("truncated svg name!");
    }

  /* find the min/max of the cpt range */

  min = cpt->segment->lsmp.val;

  max = cpt->segment->rsmp.val;
  for (seg = cpt->segment ; seg ; seg = seg->rseg)
    max = seg->rsmp.val;

  /* get the colour model */

  switch (cpt->model)
    {
    case model_rgb:
    case model_hsv:
      break;
    case model_cmyk:
      btrace("conversion of colour model %s not implemented",
	     model_name(cpt->model));
      return 1;
    default:
      btrace("unknown colour model");
      return 1;
    }

  /* convert cpt segments to svg stops*/

  for (n = 0, seg = cpt->segment ; seg ; seg = seg->rseg)
    {
      cpt_sample_t lsmp, rsmp;
      rgb_t rcol, lcol;
      cpt_fill_type_t type;

      lsmp = seg->lsmp;
      rsmp = seg->rsmp;

      if (lsmp.fill.type != rsmp.fill.type)
        {
          btrace("sorry, can't convert mixed fill types");
          return 1;
        }

      type = lsmp.fill.type;

      switch (type)
        {
        case cpt_fill_colour_rgb:
          lcol = lsmp.fill.colour.rgb;
          rcol = rsmp.fill.colour.rgb;
          break;

        case cpt_fill_colour_hsv:
          if (hsv_to_rgb(lsmp.fill.colour.hsv, &lcol) != 0 ||
              hsv_to_rgb(rsmp.fill.colour.hsv, &rcol) != 0)
            {
              btrace("failed conversion of HSV to RGB");
              return 1;
            }
          break;

        case cpt_fill_grey:
        case cpt_fill_hatch:
        case cpt_fill_file:
        case cpt_fill_empty:
	  /* fixme */
	  btrace("fill type not implemented yet");
          return 1;

        default:
          btrace("strange fill type");
          return 1;
        }

      /* always insert the left colour */

      lstop.value = 100 * (lsmp.val - min) / (max - min);
      lstop.colour = lcol;
      lstop.opacity = 1;

      if (svg_append(lstop, svg) == 0)
        m++;
      else
	{
	  btrace("error adding stop for segment %i left", n);
	  return 1;
	}

      /*
	if there is a segment to the right, and if its left
	segment is the same colour at the our right segment
	then dont insert it. Otherwise do.
      */

      if ( ! ((seg->rseg) &&
	      cpt_fill_eq(rsmp.fill, seg->rseg->lsmp.fill, cpt->model)))
	{
	  rstop.value = 100*(rsmp.val-min)/(max-min);
	  rstop.colour = rcol;
	  rstop.opacity = 1.0;

	  if (svg_append(rstop, svg) == 0) m++;
	  else
	    {
	      btrace("error adding stop for segment %i right", n);
	      return 1;
	    }
	}

      n++;
    }

  if (opt.verbose)
    printf("converted %i segments to %i stops", n, m);

  return 0;
}
