#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptutils/cpt-write.h"
#include "cptutils/btrace.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <errno.h>

static int cpt_write_stream(const cpt_t*, FILE*);
static int fprintf_cpt_aux(FILE*, char, cpt_fill_t, model_t);
static int fprintf_cpt_sample(FILE*, cpt_sample_t, model_t, const char*);
static int common_format(const cpt_t*, char*);

#define ZFMTLEN 64

int cpt_write(const cpt_t *cpt, const char *outfile)
{
  int err;

  if (outfile)
    {
      FILE *stream = fopen(outfile, "wb");

      if (stream)
	{
	  fprintf(stream, "# %s\n", outfile);
	  err = cpt_write_stream(cpt, stream);
	  fclose(stream);
	}
      else
	{
	  btrace("failed to open %s: %s", outfile, strerror(errno));
	  err = 1;
	}
    }
  else
    err = cpt_write_stream(cpt, stdout);

  return err;
}

static int cpt_write_stream(const cpt_t *cpt, FILE *stream)
{
  char zfmt[ZFMTLEN];

  if (common_format(cpt, zfmt) != 0)
    {
      btrace("failed to build z-format string");
      return 1;
    }

  /*
    This gets the string for the "COLOR MODEL" which encodes
    the colour model and interpolation model, "HSV" means
    model HSV, interpolate RGV; in GMT4 "+HSV" means model
    HSV, interpolate HSV, but in GMT5 this is replaced by
    "hsv".  Since GMT5/6 is backward compatible, it stll
    accepts "+HSV", so we keep that for now, at least until
    https://gitlab.com/jjg/cptutils/-/issues/48
  */

  const char *model;

  switch (cpt->model)
    {
    case model_rgb:
      switch (cpt->interpolate)
        {
        case model_rgb:
          model = "RGB";
          break;
        default:
          btrace("model/interpolate %s/%s",
                 model_name(cpt->model),
                 model_name(cpt->interpolate));
          return 1;
        }
      break;

    case model_hsv:
      switch (cpt->interpolate)
        {
        case model_rgb:
          model = "HSV";
          break;
        case model_hsv:
          model = "+HSV";
          break;
        default:
          btrace("model/interpolate %s/%s",
                 model_name(cpt->model),
                 model_name(cpt->interpolate));
          return 1;
        }
      break;

    case model_cmyk:
      switch (cpt->interpolate)
        {
        case model_rgb:
          model = "CMYK";
          break;
        case model_hsv:
          model = "+CMYK";
          break;
        default:
          btrace("model/interpolate %s/%s",
                 model_name(cpt->model),
                 model_name(cpt->interpolate));
          return 1;
        }
      break;

    default :
      btrace("bad model value");
      return 1;
    }

  time_t t = time(NULL);

  fprintf(stream, "# autogenerated GMT palette");
  if ((cpt->name) && (strlen(cpt->name) > 0))
    fprintf(stream, " \"%s\"", cpt->name);
  fprintf(stream, "\n");
  fprintf(stream, "# cptutils version %s, %s", VERSION, ctime(&t));
  fprintf(stream, "# COLOR_MODEL = %s\n", model);

  cpt_seg_t *seg = cpt->segment;

  /* needs to write annotation too */

  while (seg)
    {
      fprintf_cpt_sample(stream, seg->lsmp, cpt->model, zfmt);
      fprintf(stream, "  ");

      fprintf_cpt_sample(stream, seg->rsmp, cpt->model, zfmt);
      if (seg->label) fprintf(stream, " ; %s", seg->label);
      fprintf(stream, "\n");

      seg = seg->rseg;
    }

  fprintf_cpt_aux(stream, 'B', cpt->bg, cpt->model);
  fprintf_cpt_aux(stream, 'F', cpt->fg, cpt->model);
  fprintf_cpt_aux(stream, 'N', cpt->nan, cpt->model);

  return 0;
}

/*
  Given the first sample of a cpt, evaluate a suitable
  format string for all of the z-values, some heuristics
  are used here.
*/

#define MAX(a, b) (((a)>(b)) ? (a) : (b))
#define MIN(a, b) (((a)<(b)) ? (a) : (b))

#define WRATIO_MAX 100.0
#define WIDTH_MAX  1e5
#define WIDTH_MIN  1e-5
#define SIGFIG_MIN 5

static int dpused(const char *str, size_t *dpu)
{
  char *c;

  if ((c = strchr(str, '.')) == NULL) return 1;

  size_t len = strlen(c);

  if (len < 2) return 1;

  size_t i;

  for (i = len - 1 ; (i > 0) && (c[i] == '0') ; i--);

  *dpu = i;

  return 0;
}

static int common_format(const cpt_t *cpt, char *fmt)
{
  /* create array of z-values */

  size_t
    nseg = cpt_nseg(cpt),
    nstop = nseg + 1;

  if (nstop < 2)
    {
      btrace("empty cpt file");
      return 1;
    }

  double z[nstop];
  cpt_seg_t *seg = cpt->segment;

  z[0] = seg->lsmp.val;

  for (size_t i = 0 ; i < nseg ; i++)
    {
      z[i + 1] = seg->rsmp.val;
      seg = seg->rseg;
    }

  /* get min/max widths */

  double wmin = fabs(z[0]-z[1]), wmax = wmin;

  for (size_t i = 1 ; i < nstop - 1 ; i++)
    {
      double w = fabs(z[i]-z[i+1]);
      wmax = MAX(w, wmax);
      wmin = MIN(w, wmin);
    }

  /*
    if width is large/small or ratio is large then fall back to
    exponential format
  */

  if ((wmax > WRATIO_MAX * wmin) ||
      (wmax > WIDTH_MAX) ||
      (wmin < WIDTH_MIN))
    return (snprintf(fmt, ZFMTLEN, "%%7e ") >= ZFMTLEN ? 1 : 0);

  /*
    default number of dp - SIGFIG_MIN orders of magnitude less
    than wmin
  */

  int dp = ceil(SIGFIG_MIN - log10(wmin));

  dp = MAX(dp, 0);

  if (dp > 0)
    {
      /* see how many dp we acually need */

      size_t maxdpu = 0;

      for (size_t i = 0 ; i < nstop ; i++)
	{
	  if (snprintf(fmt, ZFMTLEN, "%.*f", dp, z[i]) >= ZFMTLEN)
	    return 1;

	  size_t dpu = 0;

	  if (dpused(fmt, &dpu) != 0)
	    {
	      btrace("dpused fail");
	      return 1;
	    }

	  maxdpu = MAX(maxdpu, dpu);
	}

      dp = maxdpu;
    }

  /* find the maximal width used with this dp */

  size_t maxlen = 0;

  for (size_t i = 0 ; i < nstop ; i++)
    {
      if (snprintf(fmt, ZFMTLEN, "%.*f", dp, z[i]) >= ZFMTLEN)
	return 1;

      size_t len = strlen(fmt);
      maxlen = MAX(len, maxlen);
    }

  /* generate the format string */

  if (snprintf(fmt, ZFMTLEN, "%%%zi.%df ", maxlen, dp) >= ZFMTLEN)
    return 1;

  return 0;
}

static int fprintf_cpt_fill(FILE*, cpt_fill_t, model_t);

static int fprintf_cpt_sample(FILE *stream, cpt_sample_t smp,
			      model_t model, const char *fmt)
{
  int n = 0;

  n += fprintf(stream, fmt, smp.val);
  n += fprintf_cpt_fill(stream, smp.fill, model);

  return n;
}

static int fprintf_cpt_fill(FILE *stream, cpt_fill_t fill, model_t model)
{
  int n = 0;

  switch (fill.type)
    {
    case cpt_fill_empty:
      n += fprintf(stream, "-");
      break;
    case cpt_fill_grey:
      n += fprintf(stream, "%3i", fill.grey);
      break;
    case cpt_fill_hatch:
      n += fprintf(stream, "%c%i/%i",
		   (fill.hatch.sign == 1 ? 'p' : 'P' ),
		   fill.hatch.dpi,
		   fill.hatch.n
		   );
      break;
    case cpt_fill_file:
      n += fprintf(stream, "%s", fill.file);
      break;
    case cpt_fill_colour_rgb:
      n += fprintf(stream,
                   "%3i %3i %3i",
                   fill.colour.rgb.red,
                   fill.colour.rgb.green,
                   fill.colour.rgb.blue);
      break;
    case cpt_fill_colour_hsv:
      n += fprintf(stream,
                   "%7.3f %7.5f %7.5f",
                   fill.colour.hsv.hue,
                   fill.colour.hsv.sat,
                   fill.colour.hsv.val);
      break;
    default:
      break;
    }

  return n;
}

static int fprintf_cpt_aux(FILE *stream, char c, cpt_fill_t fill, model_t model)
{
  int n = 0;

  if (fill.type == cpt_fill_empty)
    return 0;

  n += fprintf(stream, "%c ", c);
  n += fprintf_cpt_fill(stream, fill, model);
  n += fprintf(stream, "\n");

  return n;
}
