#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptutils/ggr-read.h"
#include "cptutils/path-base.h"
#include "cptutils/btrace.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define MAX_SEGMENTS 4096

#define MAX(a, b) (((a)>(b)) ? (a) : (b))
#define MIN(a, b) (((a)>(b)) ? (b) : (a))

static ggr_t* load_stream(const char*, FILE*);

ggr_t* ggr_read(const char *path)
{
  ggr_t *grad;

  if (path == NULL)
    {
      grad = load_stream(path, stdin);
    }
  else
    {
      FILE *stream = fopen(path, "rb");

      if (stream == NULL)
	{
	  btrace("failed to open %s : %s", path, strerror(errno));
	  return NULL;
	}

      grad = load_stream(path, stream);
      fclose(stream);
    }

  return grad;
}

static int load_grad(const char*, FILE*, ggr_t*);
static void warn_truncated(const char*);

static ggr_t* load_stream(const char *path, FILE *stream)
{
  char line[1024];

  if (fgets(line, 1024, stream) == NULL)
    {
      warn_truncated("at start of file");
      return NULL;
    }

  if (strncmp(line, "GIMP Gradient", 13) != 0)
    {
      btrace("file does not seem to be a GIMP gradient");
      return NULL;
    }

  ggr_t *grad = ggr_new();

  if (grad == NULL)
    return NULL;

  grad->filename = (path ? strdup(path) : strdup("<stdin>"));

  if (load_grad(path, stream, grad) != 0)
    {
      ggr_destroy(grad);
      return NULL;
    }

  return grad;
}

static void warn_truncated(const char *where)
{
  btrace("unexpected end of file %s", where);
}

static ggr_colour_r grad_hsv_type(ggr_colour_r, double, double);

static int load_grad(const char* path, FILE *stream, ggr_t *grad)
{
  char line[1024];

  if (fgets(line, 1024, stream) == NULL)
    {
      warn_truncated("after header");
      return 1;
    }

  /*
    In 1.3 gradients there is a line with the name of the
    gradient : if we find it then we use that name and read
    another line, otherwise we use the path (or the <stdin>
    string>)
  */

  if (strncmp(line, "Name:", 5) == 0)
    {
      char *s, *e;

      for (s = line+5 ; *s == ' ' ; s++);
      if ((e = strchr(s, '\n')) != NULL) *e = '\0';

      grad->name = strdup(s);

      if (fgets(line, 1024, stream) == NULL)
	{
	  warn_truncated("after segments line");
	  return 1;
	}
    }
  else
    {
      if (path == NULL)
        grad->name = strdup("cptutils-output");
      else
        grad->name = path_base(path, "*");
    }

  int num_segments = atoi(line);

  if ((num_segments < 1) || (num_segments > MAX_SEGMENTS))
    {
      btrace("invalid number of segments in %s", path);
      return 1;
    }

  ggr_segment_t *prev = NULL;

  for (int i = 0 ; i < num_segments ; i++)
    {
      ggr_segment_t *seg;
      int type, color, ect_left, ect_right;

      seg = ggr_segment_new();
      seg->prev = prev;

      if (prev)
	prev->next = seg;
      else
	grad->segments = seg;

      if (fgets(line, 1024, stream) == NULL)
	{
	  btrace("unexpected end of file at segment %i", i+1);
	  return 1;
	}

      switch (sscanf(line, "%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%d%d%d%d",
		     &(seg->left), &(seg->middle), &(seg->right),
		     &(seg->r0), &(seg->g0), &(seg->b0), &(seg->a0),
		     &(seg->r1), &(seg->g1), &(seg->b1), &(seg->a1),
		     &type, &color,
		     &ect_left, &ect_right))
	{
	case 13:

	  ect_left = ect_right = GGR_FIXED;

	  /* fallthrough */
	case 15:

	  seg->ect_left  = ect_left;
	  seg->ect_right = ect_right;

          seg->type  = (ggr_type_t)type;

	  /* determine colour model (see below) */

	  switch (color)
	    {
	    case GGR_RGB:
	    case GGR_HSV_CW:
	    case GGR_HSV_CCW:

	      seg->color = (ggr_colour_r)color;
	      break;

	    case GGR_HSV_SHORT:
	    case GGR_HSV_LONG:

	      {
		int err = 0;
		double rgb0[3], rgb1[3], hsv0[3], hsv1[3];

		rgb0[0] = seg->r0;
		rgb0[1] = seg->g0;
		rgb0[2] = seg->b0;

		err |= rgbD_to_hsvD(rgb0, hsv0);

		rgb1[0] = seg->r1;
		rgb1[1] = seg->g1;
		rgb1[2] = seg->b1;

		err |= rgbD_to_hsvD(rgb1, hsv1);

		if (err)
		  {
		    btrace("error converting rgb->hsv !");
		    return 1;
		  }

		seg->color = grad_hsv_type(color, hsv0[0], hsv1[0]);
	      }

	      break;

	    default:

	      btrace("unknown colour model (%i)", color);
	      return 1;
	    }

	  break;

	default:
	  btrace("badly formatted gradient segment %d in %s", i, path);
	  return 1;
	}
      prev = seg;
    }

  return 0;
}

/*
  this function handles an extension to the gimp gradient colour-arc
  type: the GGR_HSV_CW and GGR_HSV_CW are joined by GGR_HSV_SHORT
  which is an HSV path, either CW or CCW depending on which is
  shorter. Similarly GGR_HSV_LONG.

  This has been proposed by Neota, who provided the following function
  for its implementation. It should make it into the gimp in 2004.

  If the argument color is GGR_HSV_SHORT then this function
  returns GGR_HSV_CW or GGR_HSV_CCW depending on which hue path
  from x to y is shorter. If the argument is GGR_HSV_LONG the
  behaviour is reversed.
*/

static ggr_colour_r grad_hsv_type(ggr_colour_r color, double x, double y)
{
  double midlen, rndlen, min, max;
  int shorter, result;

  min = MIN(x, y);
  max = MAX(x, y);

  midlen = max - min;
  rndlen = min + (1.0 - max);

  /* find shorter path */

  if (rndlen < midlen)
    shorter = (max==y ? GGR_HSV_CW : GGR_HSV_CCW);
  else
    shorter = (max==y ? GGR_HSV_CCW : GGR_HSV_CW);

  switch (color)
    {
    case GGR_HSV_SHORT:
      result = shorter;
      break;
    case GGR_HSV_LONG:
      result = (shorter == GGR_HSV_CW ? GGR_HSV_CCW : GGR_HSV_CW);
      break;
    default :
      result = color;
    }

  return result;
}
