#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <cptutils/btrace.h>

#include "options.h"
#include "svgx.h"

int main(int argc, char **argv)
{
  struct gengetopt_args_info info;
  svgx_opt_t opt = {0};
  int err;

  /* use gengetopt */

  if (options(argc, argv, &info) != 0)
    {
      fprintf(stderr, "failed to parse command line\n");
      return EXIT_FAILURE;
    }

  /* check arguments & transfer to opt structure */

  opt.verbose = info.verbose_flag;
  opt.permissive = ! info.strict_flag;
  opt.name = NULL;

  if (info.all_given + info.list_given + info.select_given > 1)
    {
      fprintf(stderr, "only one of --all, --list and --select allowed\n");
      return EXIT_FAILURE;
    }

  /* job type */

  if (info.all_flag)
    opt.job = job_all;
  else if (info.list_flag)
    opt.job = job_list;
  else if (info.select_given)
    {
      opt.job = job_named;
      opt.name = info.select_arg;
    }
  else
    opt.job = job_first;

  /* output format */

  if (info.type_given)
    {
      const char *name = info.type_arg;
      struct {
        const char *name;
        int type;
      } *p, types[] = {
        {"c3g", type_css3},
        {"cpt", type_cpt},
        {"css3", type_css3},
        {"ds9", type_sao},
        {"ggr", type_ggr},
        {"gimp", type_ggr},
        {"gnuplot", type_gpt},
        {"gpt", type_gpt},
        {"inc", type_pov},
        {"jgd", type_grd3},
        {"map", type_map},
        {"pg", type_pg},
        {"pov", type_pov},
        {"png", type_png},
        {"psp", type_grd3},
        {"PspGradient", type_grd3},
        {"qgs", type_qgs},
        {"sao", type_sao},
        {"svg", type_svg},
        {"tps", type_map},
        {NULL, 0}
      };

      for (p = types ; ; p++)
	{
	  if (! p->name)
	    {
	      fprintf(stderr, "output type %s not known\n", name);

	      fprintf(stderr, "supported types:\n");
	      for (p = types ; p->name ; p++)
		fprintf(stderr, "  %s\n", p->name);
	      fprintf(stderr, "\n");

	      return EXIT_FAILURE;
	    }

	  if (strcmp(name, p->name) == 0)
	    {
	      opt.type = p->type;
	      break;
	    }
	}
    }
  else
    opt.type = type_cpt;

  /* null outfile for stdout */

  if (info.output_given)
    opt.output.file = info.output_arg;
  else
    {
      if (opt.verbose)
        {
          fprintf(stderr, "no --verbose without --output\n");
          return EXIT_FAILURE;
        }
      opt.output.file = NULL;
    }

  /* infile is required (atypically) */

  if (info.inputs_num != 1)
    {
      fprintf(stderr, "exactly one SVG input file required\n");
      return EXIT_FAILURE;
    }

  opt.input.file = info.inputs[0];

  /* fg/bg/nan for cpt output (and only that) */

  if (opt.type == type_cpt)
    {
      if (parse_rgb(info.background_arg, &(opt.format.cpt.bg)) != 0)
        {
          fprintf(stderr, "bad background %s\n", info.background_arg);
          return EXIT_FAILURE;
        }

      if (parse_rgb(info.foreground_arg, &(opt.format.cpt.fg)) != 0)
        {
          fprintf(stderr, "bad foreground %s\n", info.foreground_arg);
          return EXIT_FAILURE;
        }

      if (parse_rgb(info.nan_arg, &(opt.format.cpt.nan)) != 0)
        {
          fprintf(stderr, "bad nan colour %s\n", info.nan_arg);
          return EXIT_FAILURE;
        }
    }
  else
    {
      if (info.background_given)
        {
          fprintf(stderr, "--background is only for cpt output\n");
          return EXIT_FAILURE;
        }

      if (info.foreground_given)
        {
          fprintf(stderr, "--foreground is only for cpt output\n");
          return EXIT_FAILURE;
        }

      if (info.nan_given)
        {
          fprintf(stderr, "--nan is only for cpt output\n");
          return EXIT_FAILURE;
        }
    }

  /* colour to replace transparency if output type does not support it */

  if (info.transparency_given)
    {
      switch (opt.type)
        {
        case type_cpt:
        case type_gpt:
        case type_map:
        case type_sao:
          break;
        case type_css3:
        case type_ggr:
        case type_grd3:
        case type_pg:
        case type_png:
        case type_pov:
        case type_qgs:
        case type_svg:
          fprintf(stderr, "no --transparency for output type\n");
          return EXIT_FAILURE;
        default:
          fprintf(stderr, "unhandled output type\n");
          return EXIT_FAILURE;
        }
    }

  if (parse_rgb(info.transparency_arg, &(opt.format.alpha)) != 0)
    {
      fprintf(stderr, "bad transparency colour %s\n",
              info.transparency_arg);
      return EXIT_FAILURE;
    }

  /* parse image (preview) size for png, svg output */

  if (info.geometry_given)
    {
      switch (opt.type)
	{
	case type_png:
	case type_svg:
	  break;

	default:
	  fprintf(stderr, "no --geometry for output type\n");
	  return EXIT_FAILURE;
	}
    }

  switch (opt.type)
    {
    case type_png:

      if (sscanf(info.geometry_arg, "%zux%zu",
		 &opt.format.png.width,
		 &opt.format.png.height) != 2)
	{
	  fprintf(stderr, "bad argument \"%s\" for --geometry\n",
		  info.geometry_arg);
	  return EXIT_FAILURE;
	}
      break;

    case type_svg:

      if (info.preview_flag)
	{
	  opt.format.svg.preview.use = true;
	  if (svg_preview_geometry(info.geometry_arg,
				   &(opt.format.svg.preview)) != 0)
	    {
	      fprintf(stderr, "bad argument \"%s\" for --geometry",
		      info.geometry_arg);
	      return EXIT_FAILURE;
	    }
	}
      else
	opt.format.svg.preview.use = false;

      break;

    default:
      break;
    }

  /* say hello */

  if (opt.verbose)
    printf("This is svgx (version %s)\n", VERSION);

  /* for conversion, give details of what we will do */

  if (opt.verbose && (opt.job != job_list))
    {
      const char *tstr;

      switch (opt.type)
	{
	case type_cpt: tstr = "GMT colour palette table"; break;
	case type_css3: tstr = "CSS3 gradient"; break;
	case type_ggr: tstr = "GIMP gradient"; break;
	case type_gpt: tstr = "Gnuplot colour map"; break;
	case type_grd3: tstr = "PaintShop Pro (grd3)"; break;
        case type_map: tstr = "Tecplot map"; break;
        case type_pg: tstr = "PostGIS colour map"; break;
	case type_pov: tstr = "POV-Ray colour map"; break;
	case type_png: tstr = "png image"; break;
	case type_qgs: tstr = "QGIS style colour map"; break;
	case type_sao: tstr = "SAO (DS9) colour map"; break;
	case type_svg: tstr = "SVG gradient"; break;

	default:
	  fprintf(stderr, "unknown output format\n");
	  return EXIT_FAILURE;
	}

      printf("convert svg to %s\n", tstr);
      printf("%s format limits\n",
             (opt.permissive ? "ignoring" : "respecting"));

      if (opt.type == type_png)
	printf("png output size is %zu x %zu px\n",
	       opt.format.png.width,
	       opt.format.png.height);
      else if ((opt.type == type_svg) &&
               (opt.format.svg.preview.use))
	printf("svg preview size is %zu x %zu px\n",
	       opt.format.svg.preview.width,
	       opt.format.svg.preview.height);
    }

  /* backtrace */

  btrace_enable("svgx");

  err = svgx(&opt);

  if (err)
    {
      btrace("failed (error %i)", err);
      btrace_print_stream(stderr, BTRACE_PLAIN);

      if (info.backtrace_file_given)
	{
	  int format = BTRACE_XML;
	  if (info.backtrace_format_given)
	    {
	      format = btrace_format(info.backtrace_format_arg);
	      if (format == BTRACE_ERROR)
		{
		  fprintf(stderr, "no such backtrace format %s\n",
			  info.backtrace_format_arg);
		  return EXIT_FAILURE;
		}
	    }
	  btrace_print(info.backtrace_file_arg, format);
	}
    }

  btrace_reset();
  btrace_disable();

  if (opt.verbose)
    printf("done.\n");

  options_free(&info);

  return (err ? EXIT_FAILURE : EXIT_SUCCESS);
}
